// SPDX-License-Identifier: GPL-2.0
/*
 * ACPI Platform Firmware Runtime Update Device Driver
 *
 * Copyright (C) 2021 Intel Corporation
 * Author: Chen Yu <yu.c.chen@intel.com>
 */
#include <linux/acpi.h>
#include <linux/device.h>
#include <linux/efi.h>
#include <linux/err.h>
#include <linux/errno.h>
#include <linux/file.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/string.h>
#include <linux/uaccess.h>
#include <linux/uio.h>
#include <linux/uuid.h>
#include <uapi/linux/pfru.h>

struct pfru_device {
	guid_t uuid;
	int rev_id;
	struct device *dev;
};

static struct pfru_device pfru_dev;
static struct pfru_device *get_pfru_dev(void)
{
	return &pfru_dev;
}

static int query_capability(struct update_cap_info *cap)
{
	union acpi_object *out_obj;
	struct pfru_device *pf_dev;
	acpi_handle handle;
	int i, ret = -EINVAL;

	pf_dev = get_pfru_dev();
	handle = ACPI_HANDLE(pf_dev->dev);

	out_obj = acpi_evaluate_dsm_typed(handle, &pf_dev->uuid,
					  pf_dev->rev_id, FUNC_QUERY_UPDATE_CAP,
					  NULL, ACPI_TYPE_PACKAGE);
	if (!out_obj)
		return -EINVAL;

	for (i = 0; i < out_obj->package.count; i++) {
		union acpi_object *obj = &out_obj->package.elements[i];

		switch (i) {
		case 0:
			if (obj->type != ACPI_TYPE_INTEGER)
				goto free_acpi_buffer;
			cap->status = obj->integer.value;
			break;
		case 1:
			if (obj->type != ACPI_TYPE_INTEGER)
				goto free_acpi_buffer;
			cap->update_cap = obj->integer.value;
			break;
		case 2:
			if (obj->type != ACPI_TYPE_BUFFER)
				goto free_acpi_buffer;
			memcpy(&cap->code_type, obj->buffer.pointer,
			       obj->buffer.length);
			break;
		case 3:
			if (obj->type != ACPI_TYPE_INTEGER)
				goto free_acpi_buffer;
			cap->fw_version = obj->integer.value;
			break;
		case 4:
			if (obj->type != ACPI_TYPE_INTEGER)
				goto free_acpi_buffer;
			cap->code_rt_version = obj->integer.value;
			break;
		case 5:
			if (obj->type != ACPI_TYPE_BUFFER)
				goto free_acpi_buffer;
			memcpy(&cap->drv_type, obj->buffer.pointer,
			       obj->buffer.length);
			break;
		case 6:
			if (obj->type != ACPI_TYPE_INTEGER)
				goto free_acpi_buffer;
			cap->drv_rt_version = obj->integer.value;
			break;
		case 7:
			if (obj->type != ACPI_TYPE_INTEGER)
				goto free_acpi_buffer;
			cap->drv_svn = obj->integer.value;
			break;
		case 8:
			if (obj->type != ACPI_TYPE_BUFFER)
				goto free_acpi_buffer;
			memcpy(&cap->platform_id, obj->buffer.pointer,
			       obj->buffer.length);
			break;
		case 9:
			if (obj->type != ACPI_TYPE_BUFFER)
				goto free_acpi_buffer;
			memcpy(&cap->oem_id, obj->buffer.pointer,
			       obj->buffer.length);
			break;
		}
	}
	ret = 0;

free_acpi_buffer:
	ACPI_FREE(out_obj);

	return ret;
}

static int query_buffer(struct com_buf_info *info)
{
	union acpi_object *out_obj;
	struct pfru_device *pf_dev;
	acpi_handle handle;
	int i, ret = -EINVAL;

	pf_dev = get_pfru_dev();
	handle = ACPI_HANDLE(pf_dev->dev);

	out_obj = acpi_evaluate_dsm_typed(handle, &pf_dev->uuid,
					  pf_dev->rev_id, FUNC_QUERY_BUF,
					  NULL, ACPI_TYPE_PACKAGE);
	if (!out_obj)
		return -EINVAL;

	for (i = 0; i < out_obj->package.count; i++) {
		union acpi_object *obj = &out_obj->package.elements[i];

		switch (i) {
		case 0:
			if (obj->type != ACPI_TYPE_INTEGER)
				goto free_acpi_buffer;
			info->status = obj->integer.value;
			break;
		case 1:
			if (obj->type != ACPI_TYPE_INTEGER)
				goto free_acpi_buffer;
			info->ext_status = obj->integer.value;
			break;
		case 2:
			if (obj->type != ACPI_TYPE_INTEGER)
				goto free_acpi_buffer;
			info->addr_lo = obj->integer.value;
			break;
		case 3:
			if (obj->type != ACPI_TYPE_INTEGER)
				goto free_acpi_buffer;
			info->addr_hi = obj->integer.value;
			break;
		case 4:
			if (obj->type != ACPI_TYPE_INTEGER)
				goto free_acpi_buffer;
			info->buf_size = obj->integer.value;
			break;
		}
	}
	ret = 0;

free_acpi_buffer:
	ACPI_FREE(out_obj);

	return ret;
}

static int get_image_type(efi_manage_capsule_image_header_t *img_hdr,
			  int *type)
{
	int ret;
	guid_t code_inj_id, drv_update_id, *image_type_id;

	ret = guid_parse(PFRU_CODE_INJ_UUID, &code_inj_id);
	if (ret)
		return ret;
	ret = guid_parse(PFRU_DRV_UPDATE_UUID, &drv_update_id);
	if (ret)
		return ret;
	/* check whether this is a code injection or driver update */
	image_type_id = &img_hdr->image_type_id;
	if (guid_equal(image_type_id, &code_inj_id))
		*type = CODE_INJECT_TYPE;
	else if (guid_equal(image_type_id, &drv_update_id))
		*type = DRIVER_UPDATE_TYPE;
	else
		return -EINVAL;

	return 0;
}

/*
 * The (u64 hw_ins) was introduced in UEFI spec version 2,
 * and (u64 capsule_support) was introduced in version 3.
 * The size needs to be adjusted accordingly.
 */
static int adjust_efi_size(efi_manage_capsule_image_header_t *img_hdr,
			   int *size)
{
	int tmp_size = *size;

	tmp_size += sizeof(efi_manage_capsule_image_header_t);
	switch (img_hdr->ver) {
	case 1:
		tmp_size -= 2 * sizeof(u64);
		break;
	case 2:
		tmp_size -= sizeof(u64);
		break;
	default:
		/* only support version 1 and 2 */
		return -EINVAL;
	}
	*size = tmp_size;

	return 0;
}

/*
 * Sanity check if the capsule image has a newer version than current one.
 * Return: true if it is valid, false otherwise.
 */
static bool valid_version(const void *data, struct update_cap_info *cap)
{
	struct payload_hdr *payload_hdr;
	efi_capsule_header_t *cap_hdr;
	efi_manage_capsule_header_t *m_hdr;
	efi_manage_capsule_image_header_t *m_img_hdr;
	efi_image_auth_t *auth;
	int type, size, ret;

	cap_hdr = (efi_capsule_header_t *)data;
	size = cap_hdr->headersize;
	m_hdr = (efi_manage_capsule_header_t *)(data + size);
	/*
	 * Current data structure size plus variable array indicated
	 * by number of (emb_drv_cnt + payload_cnt)
	 */
	size += sizeof(efi_manage_capsule_header_t) +
		      (m_hdr->emb_drv_cnt + m_hdr->payload_cnt) * sizeof(u64);
	m_img_hdr = (efi_manage_capsule_image_header_t *)(data + size);

	ret = get_image_type(m_img_hdr, &type);
	if (ret)
		return false;

	ret = adjust_efi_size(m_img_hdr, &size);
	if (ret)
		return false;

	auth = (efi_image_auth_t *)(data + size);
	size += sizeof(u64) + auth->auth_info.hdr.len;
	payload_hdr = (struct payload_hdr *)(data + size);

	/* Finally, compare the version. */
	if (type == CODE_INJECT_TYPE)
		return payload_hdr->rt_ver >= cap->code_rt_version;
	else
		return payload_hdr->rt_ver >= cap->drv_rt_version;
}

static void parse_update_result(struct updated_result *result)
{
	pr_debug("Update result:\n");
	pr_debug("Status:%d\n", result->status);
	pr_debug("Extended Status:%d\n", result->ext_status);
	pr_debug("Authentication Time Low:%ld\n", result->low_auth_time);
	pr_debug("Authentication Time High:%ld\n", result->high_auth_time);
	pr_debug("Execution Time Low:%ld\n", result->low_exec_time);
	pr_debug("Execution Time High:%ld\n", result->high_exec_time);
}

static int start_acpi_update(int action)
{
	union acpi_object *out_obj, in_obj, in_buf;
	struct updated_result update_result;
	acpi_handle handle;
	struct pfru_device *pf_dev;
	int i, ret = -EINVAL;

	pf_dev = get_pfru_dev();
	memset(&in_obj, 0, sizeof(in_obj));
	memset(&in_buf, 0, sizeof(in_buf));
	in_obj.type = ACPI_TYPE_PACKAGE;
	in_obj.package.count = 1;
	in_obj.package.elements = &in_buf;
	in_buf.type = ACPI_TYPE_INTEGER;
	in_buf.integer.value = action;

	handle = ACPI_HANDLE(pf_dev->dev);
	out_obj = acpi_evaluate_dsm_typed(handle, &pf_dev->uuid,
					  pf_dev->rev_id, FUNC_START,
					  &in_obj, ACPI_TYPE_PACKAGE);
	if (!out_obj)
		return -EINVAL;

	for (i = 0; i < out_obj->package.count; i++) {
		union acpi_object *obj = &out_obj->package.elements[i];

		switch (i) {
		case 0:
			if (obj->type != ACPI_TYPE_INTEGER)
				goto free_acpi_buffer;
			update_result.status = obj->integer.value;
			break;
		case 1:
			if (obj->type != ACPI_TYPE_INTEGER)
				goto free_acpi_buffer;
			update_result.ext_status = obj->integer.value;
			break;
		case 2:
			if (obj->type != ACPI_TYPE_INTEGER)
				goto free_acpi_buffer;
			update_result.low_auth_time = obj->integer.value;
			break;
		case 3:
			if (obj->type != ACPI_TYPE_INTEGER)
				goto free_acpi_buffer;
			update_result.high_auth_time = obj->integer.value;
			break;
		case 4:
			if (obj->type != ACPI_TYPE_INTEGER)
				goto free_acpi_buffer;
			update_result.low_exec_time = obj->integer.value;
			break;
		case 5:
			if (obj->type != ACPI_TYPE_INTEGER)
				goto free_acpi_buffer;
			update_result.high_exec_time = obj->integer.value;
			break;
		}
	}
	parse_update_result(&update_result);
	ret = 0;

free_acpi_buffer:
	ACPI_FREE(out_obj);

	return ret;
}

static long pfru_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	void __user *p;
	int ret = 0, rev;
	struct pfru_device *pf_dev;

	pf_dev = get_pfru_dev();

	p = (void __user *)arg;

	switch (cmd) {
	case PFRU_IOC_SET_REV:
		if (copy_from_user(&rev, p, sizeof(unsigned int)))
			return -EFAULT;
		if (!valid_revid(rev))
			return -EFAULT;
		pf_dev->rev_id = rev;
		break;
	case PFRU_IOC_STAGE:
		ret = start_acpi_update(START_STAGE);
		if (ret)
			return ret;
		break;
	case PFRU_IOC_ACTIVATE:
		ret = start_acpi_update(START_ACTIVATE);
		if (ret)
			return ret;
		break;
	case PFRU_IOC_STAGE_ACTIVATE:
		ret = start_acpi_update(START_STAGE_ACTIVATE);
		if (ret)
			return ret;
		break;
	default:
		ret = -ENOIOCTLCMD;
		break;
	}

	return ret;
}

#ifdef CONFIG_COMPAT
static long compat_pfru_ioctl(struct file *filep, unsigned int cmd,
			      unsigned long arg)
{
	return pfru_ioctl(filep, cmd, arg);
}
#endif

static int pfru_open(struct inode *inode, struct file *file)
{
	return capable(CAP_SYS_RAWIO) ? stream_open(inode, file) : -EPERM;
}

static ssize_t pfru_write(struct file *file, const char __user *buf,
			  size_t len, loff_t *ppos)
{
	struct update_cap_info cap;
	struct com_buf_info info;
	phys_addr_t phy_addr;
	struct iov_iter iter;
	struct iovec iov;
	char *buf_ptr;
	int ret;

	ret = query_buffer(&info);
	if (ret)
		return ret;

	if (len > info.buf_size)
		return -EINVAL;

	iov.iov_base = (void __user *)buf;
	iov.iov_len = len;
	iov_iter_init(&iter, WRITE, &iov, 1, len);

	/* map the communication buffer */
	phy_addr = (phys_addr_t)(info.addr_lo | (info.addr_hi << 32));
	buf_ptr = memremap(phy_addr, info.buf_size, MEMREMAP_WB);
	if (!buf_ptr)
		return -ENOMEM;
	if (!copy_from_iter_full(buf_ptr, len, &iter)) {
		pr_err("error! could not read capsule file\n");
		ret = -EINVAL;
		goto unmap;
	}

	/* Check if the capsule header has a valid version number. */
	ret = query_capability(&cap);
	if (ret)
		goto unmap;

	if (cap.status != DSM_SUCCEED) {
		ret = -EBUSY;
		goto unmap;
	}
	if (!valid_version(buf_ptr, &cap)) {
		ret = -EINVAL;
		goto unmap;
	}
	ret = 0;
unmap:
	memunmap(buf_ptr);

	return ret ?: len;
}

static ssize_t pfru_read(struct file *filp, char __user *ubuf,
			 size_t size, loff_t *off)
{
	struct update_cap_info cap;
	int ret;

	ret = query_capability(&cap);
	if (ret)
		return ret;

	size = min_t(size_t, size, sizeof(cap));

	if (copy_to_user(ubuf, &cap, size))
		return -EFAULT;

	return size;
}

static const struct file_operations acpi_pfru_fops = {
	.owner		= THIS_MODULE,
	.write		= pfru_write,
	.read		= pfru_read,
	.open		= pfru_open,
	.unlocked_ioctl = pfru_ioctl,
#ifdef CONFIG_COMPAT
	.compat_ioctl	= compat_pfru_ioctl,
#endif
	.llseek		= noop_llseek,
};

static struct miscdevice pfru_misc_dev = {
	.minor = MISC_DYNAMIC_MINOR,
	.name = "pfru_update",
	.nodename = "pfru/update",
	.fops = &acpi_pfru_fops,
};

static int acpi_pfru_remove(struct platform_device *pdev)
{
	misc_deregister(&pfru_misc_dev);

	return 0;
}

static int acpi_pfru_probe(struct platform_device *pdev)
{
	acpi_handle handle;
	struct pfru_device *pf_dev;
	int ret;

	pf_dev = get_pfru_dev();

	ret = guid_parse(PFRU_UUID, &pf_dev->uuid);
	if (ret)
		return ret;
	/* default rev id is 1 */
	pf_dev->rev_id = 1;
	pf_dev->dev = &pdev->dev;
	handle = ACPI_HANDLE(pf_dev->dev);
	if (!acpi_has_method(handle, "_DSM")) {
		pr_err("Missing _DSM\n");
		return -ENODEV;
	}

	ret = misc_register(&pfru_misc_dev);
	if (ret)
		return ret;

	return 0;
}

static const struct acpi_device_id acpi_pfru_ids[] = {
	{"INTC1080", 0},
	{}
};
MODULE_DEVICE_TABLE(acpi, acpi_pfru_ids);

static struct platform_driver acpi_pfru_driver = {
	.driver = {
		.name = "pfru_update",
		.acpi_match_table = acpi_pfru_ids,
	},
	.probe = acpi_pfru_probe,
	.remove = acpi_pfru_remove,
};
module_platform_driver(acpi_pfru_driver);

MODULE_DESCRIPTION("Platform Firmware Runtime Update device driver");
MODULE_LICENSE("GPL v2");
